import os

from ._internal import DEFAULT, defaultarguments, DoreahConfig


# defaultextension	files with this extension will be regarded as valid files. can be overwritten per request.
# comments			whether files may include commenting (indicated by #)
# multitab			whether fields can be separated by multiple tabs (this makes empty fields impossible except when trailing)
config = DoreahConfig("tsv",
	defaultextension=".tsv",
	comments=True,
	multitab=True
)


@defaultarguments(config,comments="comments",multitab="multitab")
def parse(filename,*args,comments=DEFAULT,multitab=DEFAULT):
	"""Parses a tsv-like file.

	:param string filename: File to be parsed
	:param string args: Data types to be read
	:param boolean comments: Whether the file may contain comments
	:param boolean multitab: Whether values can be separated by an arbitrary amount of tabs
	:return: List of entries, one entry per line, each entry a list of fields"""

	if not os.path.exists(filename):
		filename = filename + config["defaultextension"]

	f = open(filename)

	result = []
	for l in [l for l in f if (not l.startswith("#")) and (not l.strip()=="")]:
		l = l.replace("\n","")

		# if the file allows comments, we need to replace the escape sequence and properly stop parsing for inline comments
		if comments:
			l = l.split("#")[0]
			l = l.replace(r"\num","#")
			l = l.replace(r"\hashtag","#")

		# we either allow multiple tabs, or we don't (in which case empty fields are possible)
		if multitab:
			data = list(filter(None,l.split("\t")))
		else:
			data = list(l.split("\t"))

		entry = [] * len(args)
		for i in range(len(args)):
			if args[i] in ["list","ls","array"]:
				try:
					entry.append(data[i].split("␟"))
				except:
					entry.append([])
			elif args[i] in ["string","str","text"]:
				try:
					entry.append(data[i])
				except:
					entry.append("")
			elif args[i] in ["int","integer","num","number"]:
				try:
					entry.append(int(data[i]))
				except:
					entry.append(0)
			elif args[i] in ["bool","boolean"]:
				try:
					entry.append((data[i].lower() in ["true","yes","1","y"]))
				except:
					entry.append(False)
			else:
				raise TypeError()

		result.append(entry)

	f.close()
	return result

@defaultarguments(config,extension="defaultextension")
def parse_all(path,*args,extension=DEFAULT,**kwargs):
	"""Parses all valid files in a specific directory

	:param string path: Directory to search for files
	:param string args: Data types to be read
	:param string extension: Only parse files ending with this string
	:param kwargs: Parse settings to be applied, see :func:`parse`"""

	result = []
	for f in os.listdir(path + "/"):
		if (f.endswith(extension)): # use "" if all files are valid
			result += parse(path + "/" + f,*args,**kwargs)

	return result



def create(filename):
	"""Creates a file if it doesn't already exist

	:param string filename: Full filename"""

	if not os.path.exists(filename):
		open(filename,"w").close()

@defaultarguments(config,comments="comments")
def add_entry(filename,entry,comments=DEFAULT):
	"""Adds an entry to a tsv-like file

	:param string filename: Full filename
	:param iterable entry: List or tuple of fields
	:param boolean comments: Whether the file can contain comments"""

	create(filename)
	# remove all tabs and create tab-separated string
	line = "\t".join([str(field).replace("\t"," ") for field in entry])

	# replace comment symbol
	if comments: line = line.replace("#",r"\num")

	with open(filename,"a") as f:
		f.write(line + "\n")

@defaultarguments(config,comments="comments")
def add_entries(filename,entries,comments=DEFAULT):
	"""Adds several entries to a tsv-like file

	:param string filename: Full filename
	:param iterable entries: List or tuple of entries, each a list of tuple of fields
	:param boolean comments: Whether the file can contain comments"""

	create(filename)

	with open(filename,"a") as f:
		for entry in entries:
			line = "\t".join([str(field).replace("\t"," ") for field in entry])
			if comments: line = line.replace("#",r"\num")
			f.write(line + "\n")



import os
from functools import wraps
import yaml


# This is purely for documentation
class Default:
	def __repr__(self):
		return u"<Module default>"
DEFAULT = Default()

## decorator to set default arguments that are only evaluated at runtime
def defaultarguments(defaultdict,**defaultargs):
	def decorator(func):	#actual decorator function
		@wraps(func)
		def newfunc(*args,**kwargs):
			realargs = {}

			# evaluate default values at runtime
			for k in defaultargs:
				realargs[k] = defaultdict[defaultargs[k]]

			# overwrite given arguments of function
			realargs.update(kwargs)

			# execute function with correct arguments
			return func(*args,**realargs)

		return newfunc

	return decorator


## opens file, creates it if necessary (including all folders)
def gopen(filepath,mode):
	directory = os.path.dirname(filepath)
	os.makedirs(directory, exist_ok=True)

	return open(filepath,mode)





class DoreahConfig:

	def __init__(self,module,**defaults):
		self.module = module
		self.configuration = {}
		self.configuration.update(defaults)
		self._readfile()
		self._readpreconfig()


	def _readfile(self):
		try:
			with open(".doreah","r") as fil:
				s = yaml.safe_load(fil).get(self.module)
			if s is not None: self.configuration.update(s)
		except OSError:
			pass
		except:
			print("Doreah could not read its configuration file. Your application is likely not up to date and uses the old doreah format!")

	def _readpreconfig(self):
		s = _preconfig.get(self.module)
		if s is not None: self.configuration.update(s)


	# set configuration (exposes function to configure module)
	def __call__(self,**kwargs):
		self.configuration.update(kwargs)


	def __getitem__(self,key):
		return self.configuration[key]

	def __repr__(self):
		return "Configuration for " + self.module + ": " + repr(self.configuration)

_preconfig = {}

def config(**modules):
	global _preconfig
	_preconfig.update(modules)



from ._internal import DEFAULT, defaultarguments, DoreahConfig
from .io import col
import zipfile
import os
import urllib.request
import shutil
import distutils.dir_util
import sys
import subprocess
import signal
#import inspect


config = DoreahConfig("control")


def cmd_handle(shortcuts):
	cmd = sys.argv[1:]
	args = []
	kwargs = {}

	while len(cmd) > 0:
		if cmd[0].startswith("-") and cmd[0][1:] in shortcuts:
			kwargs[shortcuts[cmd[0][1:]]] = cmd[1]
			cmd = cmd[2:]
		elif cmd[0].startswith("--"):
			kwargs[cmd[0][2:]] = cmd[1]
			cmd = cmd[2:]
		else:
			args.append(cmd[0])
			cmd = cmd[1:]

	return args,kwargs

def mainfunction(shortcuts,shield=False):
	def decorator(func):
		# define a wrapper function that takes no args itself,
		# but reads them from console and passes them down
		def wrapper():
			args,kwargs = cmd_handle(shortcuts)
			for var in kwargs:
				if var in func.__annotations__:
					kwargs[var] = func.__annotations__[var](kwargs[var])
			return func(*args,**kwargs)

		# if this module is the script, call that function immediately
		if func.__module__ == "__main__":
			wrapper()

		# either return the original function if used with arguments in other places,
		# or return the wrapper if the function is called from command line through
		# some other means (console script)
		return wrapper if shield else func

	return decorator


class Controller:
	def __init__(self,pkgname,processname=None,prettyname=None):
		self.pkgname = pkgname
		self.processname = processname if processname is not None else self.pkgname
		self.prettyname = prettyname if prettyname is not None else self.processname.capitalize()

		self.actions = {
			"start":self.start,
			"restart":self.restart,
			"stop":self.stop
		}


	def getInstance(self):
		try:
			output = subprocess.check_output(["pidof",self.processname])
			pid = int(output)
			return pid
		except:
			return None
	def is_running(self):
		return getInstance() is not None

	def start(self):
		try:
			p = subprocess.Popen(["python3","-m",self.pkgname + ".main"],stdout=subprocess.DEVNULL,stderr=subprocess.DEVNULL)
			print(col["green"](self.prettyname + " started!") + " PID: " + str(p.pid))
		except:
			print("Error while starting " + self.prettyname + ".")
			return False

	def restart(self):
		wasrunning = stop()
		start()
		return wasrunning

	def stop(self):
		pid = getInstance()
		if pid is None:
			print(self.prettyname + " is not running")
			return False
		else:
			os.kill(pid,signal.SIGTERM)
			print(self.prettyname + " stopped! PID: " + str(pid))
			return True
